from invoiced.client import Client  # noqa
from invoiced.version import VERSION  # noqa
from invoiced.objects import (  # noqa
    Attachment,
    BankAccount,
    Card,
    CatalogItem,
    Contact,
    Coupon,
    CreditNote,
    Customer,
    Email,
    Estimate,
    Event,
    File,
    Invoice,
    Letter,
    LineItem,
    Note,
    PaymentPlan,
    PaymentSource,
    Plan,
    Subscription,
    Task,
    TaxRate,
    TextMessage,
    Transaction)
from invoiced.operations import List  # noqa



class ErrorBase(Exception):
    def __init__(self, message=None, status_code=None, error=None):
        super(ErrorBase, self).__init__(message)

        self._message = message
        self.status_code = status_code
        self.error = error

    def __str__(self):
        return "({0}): {1}".format(self.status_code, self._message)


class ApiConnectionError(ErrorBase):
    pass


class ApiError(ErrorBase):
    pass


class InvalidRequestError(ErrorBase):
    pass


class RateLimitError(ErrorBase):
    pass


class AuthenticationError(ErrorBase):
    pass



from inflection import (pluralize, underscore)
from invoiced import util
import json
import re


class InvoicedObject(dict):

    def __init__(self, client, id=None, values={}):
        super(InvoicedObject, self).__init__()

        self._client = client

        # generate endpoint based on class name
        class_name = self.__class__.__name__
        self._endpoint_base = ''
        self._endpoint = '/' + pluralize(underscore(class_name)).lower()

        if id:
            self._endpoint = self._endpoint + '/' + str(id)
            self._unsaved = set()
            values.update({'id': id})
            super(InvoicedObject, self).update(values)

    def set_endpoint_base(self, base):
        self._endpoint_base = base

        return self

    def endpoint_base(self):
        return self._endpoint_base

    def endpoint(self):
        return self._endpoint_base + self._endpoint

    def retrieve(self, id, opts={}):
        if not id:
            raise ValueError("Missing ID.")

        response = self._client.request('GET',
                                        self.endpoint()+"/"+str(id),
                                        opts)

        return util.convert_to_object(self, response['body'])

    def refresh_from(self, values):
        self._unsaved = set()
        self.clear()

        for k, v in values.items():
            super(InvoicedObject, self).__setitem__(k, v)

    def update(self, update_dict):
        for k in update_dict:
            self._unsaved.add(k)

        return super(InvoicedObject, self).update(update_dict)

    def __setattr__(self, k, v):
        if k[0] == '_' or k in self.__dict__:
            return super(InvoicedObject, self).__setattr__(k, v)
        else:
            self[k] = v

    def __getattr__(self, k):
        if k[0] == '_':
            raise AttributeError(k)

        try:
            return self[k]
        except KeyError as err:
            raise AttributeError(*err.args)

    def __delattr__(self, k):
        if k[0] == '_' or k in self.__dict__:
            return super(InvoicedObject, self).__delattr__(k)
        else:
            del self[k]

    def __setitem__(self, k, v):
        if v == "":
            raise ValueError(
                "You cannot set %s to an empty string. "
                "We interpret empty strings as None in requests."
                "You may set %s.%s = None to delete the property" % (
                    k, str(self), k))

        super(InvoicedObject, self).__setitem__(k, v)

        self._unsaved.add(k)

    def __delitem__(self, k):
        super(InvoicedObject, self).__delitem__(k)

        self._unsaved.remove(k)

    def __repr__(self):
        ident_parts = [type(self).__name__]

        if self.get('id'):
            ident_parts.append('id=%s' % str(self.get('id')))

        return '<%s at %s> JSON: %s' % (
            ' '.join(ident_parts), hex(id(self)), str(self))

    def __str__(self):
        return json.dumps(self, sort_keys=True, indent=2)


class PaymentSourceObject(InvoicedObject):
    pass

class CreateableObject(InvoicedObject):

    def create(self, idempotency_key=None, **params):
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', self.endpoint(), params, opts)

        return util.convert_to_object(self, response['body'])


class DeleteableObject(InvoicedObject):

    def delete(self):
        response = self._client.request('DELETE', self.endpoint())

        if response['code'] == 204:
            self.refresh_from({'id': self.id})
        elif response['code'] == 200:
            # update the local values with the response
            self.refresh_from(response['body'])

        return response['code'] == 204 or response['code'] == 200


class ListableObject(InvoicedObject):

    def list(self, **opts):
        response = self._client.request('GET', self.endpoint(), opts)

        # build objects
        objects = util.build_objects(self, response['body'])

        # store the metadata from the list operation
        metadata = List(response['headers']['link'],
                        response['headers']['x-total-count'])

        return objects, metadata


class UpdateableObject(InvoicedObject):

    def save(self, idempotency_key=None, **params):
        update = {}
        for k in self._unsaved:
            update[k] = self[k]

        update.update(params)

        # perform the update if there are any changes
        if len(update) > 0:
            opts = {'idempotency_key': idempotency_key}
            response = self._client.request('PATCH',
                                            self.endpoint(),
                                            update,
                                            opts)

            # update the local values with the response
            self.refresh_from(response['body'])

            return response['code'] == 200

        return False


class List:
    def __init__(self, link_header, total_count):
        self.links = self.parse_link_header(link_header)
        self.total_count = int(total_count)

    def parse_link_header(self, header):
        links = {}

        # Parse each part into a named link
        for part in header.split(','):
            section = part.split(';')
            url = re.search('<(.*)>', section[0]).group(1)
            name = re.search('rel="(.*)"', section[1]).group(1)
            links[name] = url

        return links



import unittest
import invoiced
import responses
import json
from requests.exceptions import HTTPError, ConnectTimeout, ReadTimeout


class TestClient(unittest.TestCase):

    def test_new_client(self):
        client = invoiced.Client('api_key')

        self.assertEqual('api_key', client.api_key)
        self.assertEqual('https://api.invoiced.com', client.api_url)
        self.assertIsInstance(client.Customer, invoiced.Customer)
        self.assertIsInstance(client.Invoice, invoiced.Invoice)
        self.assertIsInstance(client.Transaction, invoiced.Transaction)
        self.assertIsInstance(client.Subscription, invoiced.Subscription)

    def test_new_client_sandbox(self):
        client = invoiced.Client('api_key', True)

        self.assertEqual('api_key', client.api_key)
        self.assertEqual('https://api.sandbox.invoiced.com', client.api_url)
        self.assertIsInstance(client.Customer, invoiced.Customer)
        self.assertIsInstance(client.Invoice, invoiced.Invoice)
        self.assertIsInstance(client.Transaction, invoiced.Transaction)
        self.assertIsInstance(client.Subscription, invoiced.Subscription)

    @responses.activate
    def test_get_request(self):
        def request_callback(request):
            # verify headers
            self.assertEqual(request.headers['authorization'], "Basic dGVzdDo=")  # noqa
            self.assertEqual(request.headers['content-type'], "application/json")  # noqa
            self.assertEqual(request.headers['user-agent'], "Invoiced Python/" + invoiced.VERSION)  # noqa

            headers = {'Header': 'test', 'Content-Type': 'application/json'}
            body = {'test': True}
            return (200, headers, json.dumps(body))

        responses.add_callback(responses.GET,
                               'https://api.invoiced.com/invoices?filter[levels]=work&test=property',  # noqa
                               match_querystring=True,
                               callback=request_callback)

        client = invoiced.Client('test')

        params = {
            'test': 'property',
            'filter': {
                'levels': 'work'
            }
        }
        response = client.request('GET', '/invoices', params)

        expectedResponse = {
            'code': 200,
            'headers': {
                'Header': 'test',
                'Content-Type': 'application/json'
            },
            'body': {
                'test': True
            }
        }
        self.assertEqual(response, expectedResponse)

    @responses.activate
    def test_post_request(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      status=204,
                      adding_headers={'Header': 'test'})

        client = invoiced.Client('test')

        params = {
            "test": "property"
        }
        response = client.request("POST", "/invoices", params)

        expectedResponse = {
            'code': 204,
            'headers': {
                'Header': 'test',
                'Content-Type': 'text/plain'
            },
            'body': None
        }
        self.assertEqual(response, expectedResponse)

    @responses.activate
    def test_post_idempotent_request(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      status=204,
                      adding_headers={'Header': 'test'})

        client = invoiced.Client('test')

        params = {
            "test": "property"
        }
        opts = {
            "idempotency_key": "a random value"
        }
        response = client.request("POST", "/invoices", params, opts)

        expectedResponse = {
            'code': 204,
            'headers': {
                'Header': 'test',
                'Content-Type': 'text/plain'
            },
            'body': None
        }
        self.assertEqual(response, expectedResponse)

    @responses.activate
    def test_request_exception(self):
        responses.add(responses.GET,
                      'https://api.invoiced.com/invoices',
                      body=HTTPError('Something went wrong'))

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.ApiConnectionError):
            client.request("GET", "/invoices")

    @responses.activate
    def test_connect_timeout(self):
        responses.add(responses.GET,
                      'https://api.invoiced.com/invoices',
                      body=ConnectTimeout('Something went wrong'))

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.ApiConnectionError):
            client.request("GET", "/invoices")

    @responses.activate
    def test_read_timeout(self):
        responses.add(responses.GET,
                      'https://api.invoiced.com/invoices',
                      body=ReadTimeout('Something went wrong'))

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.ApiConnectionError):
            client.request("GET", "/invoices")

    @responses.activate
    def test_invalid_request_error(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      json={'message': 'error'},
                      status=400)

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.InvalidRequestError):
            client.request("POST", "/invoices")

    @responses.activate
    def test_rate_limit_error(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      json={'message': 'error'},
                      status=429)

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.RateLimitError):
            client.request("POST", "/invoices")

    @responses.activate
    def test_authentication_error(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      json={'message': 'error'},
                      status=401)

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.AuthenticationError):
            client.request("POST", "/invoices")

    @responses.activate
    def test_api_error(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      json={'message': 'error'},
                      status=500)

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.ApiError):
            client.request("POST", "/invoices")

    @responses.activate
    def test_api_error_invalid_json(self):
        responses.add(responses.POST,
                      'https://api.invoiced.com/invoices',
                      body='not valid json',
                      content_type='application/json',
                      status=400)

        client = invoiced.Client('test')

        with self.assertRaises(invoiced.errors.ApiError):
            client.request("POST", "/invoices")


if __name__ == '__main__':
    unittest.main()



import unittest
import invoiced
from invoiced import util


class TestUtil(unittest.TestCase):

    def setUp(self):
        self.client = invoiced.Client('api_key')

    def test_convert_to_object(self):
        customer = {'id': 100}
        obj = util.convert_to_object(self.client.Customer, customer)

        self.assertIsInstance(obj, invoiced.Customer)
        self.assertEqual(obj.id, 100)

    def test_build_objects(self):
        customers = [
            {'id': 100},
            {'id': 101}
        ]

        objects = util.build_objects(self.client.Customer, customers)
        self.assertEqual(2, len(objects))
        self.assertIsInstance(objects[0], invoiced.Customer)
        self.assertIsInstance(objects[1], invoiced.Customer)
        self.assertEqual(objects[0].id, 100)
        self.assertEqual(objects[1].id, 101)

    def test_uri_encode(self):
        params = {
            "test": "property",
            "filter": {
                "levels": "work",
                "nesting": {
                    "works": True
                }
            },
            "array": [
                "should",
                {"also": True},
                ["work"]
            ]
        }

        encoded = util.uri_encode(params)
        self.assertEqual(encoded, "array[]=should&array[][also]=True&array[]=work&filter[levels]=work&filter[nesting][works]=True&test=property")  # noqa

if __name__ == '__main__':
    unittest.main()



import os
import unittest


def all():
    path = os.path.dirname(os.path.realpath(__file__))
    print(path)
    return unittest.defaultTestLoader.discover(path)


def objects():
    path = os.path.dirname(os.path.realpath(__file__))
    return unittest.defaultTestLoader.discover(
        os.path.join(path, "objects"))



import urllib

def build_objects(_class, objects):
    return list(map(lambda obj: convert_to_object(_class, obj), objects))


def convert_to_object(_class, values):
    c = _class.__class__

    obj = None

    # handle special cases w/ Card, BankAccount
    if 'object' in values.keys():
        if values['object'] == 'card':
            from invoiced import Card
            obj = Card(_class._client, values['id'], values)
        elif values['object'] == 'bank_account':
            from invoiced import BankAccount
            obj = BankAccount(_class._client, values['id'], values)

    if obj is None:
        obj = c(_class._client, values['id'], values)

    obj.set_endpoint_base(_class.endpoint_base())

    return obj


def uri_encode(params):
    return '&'.join(map(lambda pair: "%s=%s" % pair, _flatten_params(params)))


def _url_encode(value):
    return urllib.parse.quote(str(value))


def _flatten_params(params, parent_key=None):
    result = []
    for key in sorted(params.keys()):
        value = params[key]
        key = _url_encode(key)
        calculated_key = parent_key+"["+key+"]" if parent_key else key
        if type(value) is dict:
            result += _flatten_params(value, calculated_key)
        elif type(value) is list:
            result += _flatten_params_array(value, calculated_key)
        else:
            result.append((calculated_key, _url_encode(value)))

    return result


def _flatten_params_array(value, calculated_key):
    result = []
    for elem in value:
        if type(elem) is dict:
            result += _flatten_params(elem, calculated_key+"[]")
        elif type(elem) is list:
            result += _flatten_params_array(elem, calculated_key)
        else:
            result.append((calculated_key+"[]", _url_encode(elem)))

    return result


from invoiced.objects import (
    CatalogItem,
    Coupon,
    CreditNote,
    Customer,
    Estimate,
    Event,
    File,
    Invoice,
    Letter,
    Note,
    Plan,
    Task,
    TaxRate,
    TextMessage,
    Transaction,
    Subscription)
from invoiced import errors, util, version
import json
import requests
from requests.auth import HTTPBasicAuth


class Client(object):

    ApiBase = 'https://api.invoiced.com'
    ApiBaseSandbox = 'https://api.sandbox.invoiced.com'

    ConnectTimeout = 30
    ReadTimeout = 60

    def __init__(self, api_key, sandbox=False):
        self.api_key = api_key
        self.sandbox = sandbox
        self.api_url = self.ApiBaseSandbox if sandbox else self.ApiBase

        # Object endpoints
        self.CatalogItem = CatalogItem(self)
        self.Coupon = Coupon(self)
        self.CreditNote = CreditNote(self)
        self.Customer = Customer(self)
        self.Estimate = Estimate(self)
        self.Event = Event(self)
        self.File = File(self)
        self.Invoice = Invoice(self)
        self.Note = Note(self)
        self.Plan = Plan(self)
        self.Subscription = Subscription(self)
        self.Task = Task(self)
        self.TaxRate = TaxRate(self)
        self.Transaction = Transaction(self)

    def request(self, method, endpoint, params={}, opts={}):
        url = self.api_url + endpoint

        # These methods don't have a request body
        if method in ('GET', 'HEAD', 'DELETE'):
            payload = None
            # Make params into GET parameters
            if len(params) > 0:
                url = url + "?" + util.uri_encode(params)
        # Otherwise, encode request body to JSON
        else:
            payload = json.dumps(params, separators=(',', ':'))

        try:
            resp = requests.request(method, url,
                                    headers=self.build_headers(opts),
                                    data=payload,
                                    auth=HTTPBasicAuth(self.api_key, ''),
                                    timeout=(self.ConnectTimeout,
                                             self.ReadTimeout))

            if (resp.status_code >= 400):
                self.handle_api_error(resp)
        except requests.exceptions.RequestException as e:
            self.handle_network_error(e)

        return self.parse(resp)

    def build_headers(self, opts):
        headers = {
            'content-type': "application/json",
            'user-agent': "Invoiced Python/"+version.VERSION
        }

        # idempotency keys
        if "idempotency_key" in opts and opts["idempotency_key"]:
            headers["idempotency-key"] = opts["idempotency_key"]

        return headers

    def parse(self, response):
        if response.status_code == 204:
            parsed_response = None
        else:
            parsed_response = json.loads(response.text)

        return {
            'code': response.status_code,
            'headers': response.headers,
            'body': parsed_response
        }

    def handle_api_error(self, response):
        try:
            error = json.loads(response.text)
        except:
            raise self.general_api_error(response.status_code, response.text)

        if response.status_code in (400, 403, 404):
            raise self.invalid_request_error(error, response)
        elif response.status_code == 401:
            raise self.authentication_error(error, response)
        elif response.status_code == 429:
            raise self.rate_limit_error(error, response)
        else:
            raise self.api_error(error, response)

    def handle_network_error(self, error):
        if isinstance(error, requests.exceptions.ConnectTimeout):
            message = ("Timed out while connecting to Invoiced. "
                       "Please check your internet connection or "
                       "status.invoiced.com for service outages.")
        elif isinstance(error, requests.exceptions.ReadTimeout):
            message = "The request timed out reading data from the server."
        else:
            message = ("There was an error connecting to "
                       "Invoiced. Please check your internet "
                       "connection or status.invoiced.com "
                       "for service outages. The reason was: "
                       + str(error))

        raise errors.ApiConnectionError(message)

    def authentication_error(self, error, response):
        return errors.AuthenticationError(error["message"],
                                          response.status_code,
                                          error)

    def invalid_request_error(self, error, response):
        return errors.InvalidRequestError(error["message"],
                                          response.status_code,
                                          error)

    def rate_limit_error(self, error, response):
        return errors.RateLimitError(error["message"],
                                     response.status_code,
                                     error)

    def api_error(self, error, response):
        return errors.ApiError(error["message"], response.status_code, error)

    def general_api_error(self, code, body):
        return errors.ApiError("API Error " + str(code) + " - " +
                               str(body), code)



from invoiced.errors import ApiError
import unittest


class TestError(unittest.TestCase):

    def test_error(self):
        error = ApiError("ERROR!", 500)
        self.assertEqual("(500): ERROR!", str(error))

if __name__ == '__main__':
    unittest.main()



import os
import sys
import warnings

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

path, script = os.path.split(sys.argv[0])
os.chdir(os.path.abspath(path))

install_requires = []

if sys.version_info < (3, 0):
    warnings.warn(
        'Python 2 is not supported.',
        DeprecationWarning)

sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'invoiced'))
from version import VERSION

with open('LONG_DESCRIPTION.rst') as f:
    long_description = f.read()

setup(
    name='invoiced',
    version=VERSION,
    description='Invoiced client library',
    long_description=long_description,
    author='Invoiced',
    author_email='support@invoiced.com',
    url='https://github.com/invoiced/invoiced-python',
    packages=['invoiced', 'invoiced.test'],
    install_requires=[
        'requests >= 2.0.0',
        'inflection == 0.3.1'
    ],
    extras_require={
        'requests': ['security>=2.0.0']
    },
    test_suite='invoiced.test.all',
    tests_require=['responses == 0.5.0'],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ])



from invoiced.operations import (
    InvoicedObject,
    PaymentSourceObject,
    CreateableObject,
    DeleteableObject,
    ListableObject,
    UpdateableObject,
    List)
from invoiced import util


class Attachment(InvoicedObject):
    pass

class BankAccount(PaymentSourceObject, DeleteableObject):
    pass

class Card(PaymentSourceObject, DeleteableObject):
    pass

class CatalogItem(CreateableObject, DeleteableObject, ListableObject,
                  UpdateableObject):
    pass

class Contact(CreateableObject, DeleteableObject, ListableObject,
              UpdateableObject):
    pass

class Coupon(CreateableObject, DeleteableObject, ListableObject,
              UpdateableObject):
    pass

class CreditNote(CreateableObject, DeleteableObject, ListableObject,
                 UpdateableObject):

    def send(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/emails"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build email objects
        email = Email(self._client)
        return util.build_objects(email, response['body'])

    def attachments(self, **params):
        response = self._client.request('GET',
                                        self.endpoint()+"/attachments",
                                        params)

        # ensure each attachment has an ID
        body = response['body']
        for attachment in body:
            if 'id' not in attachment:
                attachment['id'] = attachment['file']['id']

        # build attachment objects
        attachment = Attachment(self._client)
        attachments = util.build_objects(attachment, body)

        # store the metadata from the list operation
        metadata = List(response['headers']['link'],
                        response['headers']['x-total-count'])

        return attachments, metadata

    def void(self):
        endpoint = self.endpoint()+'/void'
        response = self._client.request('POST', endpoint)

        # update the local values with the response
        self.refresh_from(response['body'])

        return response['code'] == 200


class Customer(CreateableObject, DeleteableObject, ListableObject,
               UpdateableObject):

    def send_statement(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/emails"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build email objects
        email = Email(self._client)
        return util.build_objects(email, response['body'])

    def send_statement_sms(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/text_messages"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build text message objects
        text_message = TextMessage(self._client)
        return util.build_objects(text_message, response['body'])

    def send_statement_letter(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/letters"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build letter objects
        letter = Letter(self._client)
        return util.build_objects(letter, response['body'])

    def balance(self):
        endpoint = self.endpoint()+"/balance"
        response = self._client.request('GET', endpoint)

        return response['body']

    def contacts(self):
        contact = Contact(self._client)
        contact.set_endpoint_base(self.endpoint())

        return contact

    def list_notes(self):
        note = Note(self._client)
        note.set_endpoint_base(self.endpoint())

        return note

    def line_items(self):
        line = LineItem(self._client)
        line.set_endpoint_base(self.endpoint())

        return line

    def payment_sources(self):
        source = PaymentSource(self._client)
        source.set_endpoint_base(self.endpoint())

        return source

    def invoice(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/invoices"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build invoice object
        invoice = Invoice(self._client)
        return util.convert_to_object(invoice, response['body'])

    def consolidate_invoices(self, **params):
        endpoint = self.endpoint()+"/consolidate_invoices"
        response = self._client.request('POST', endpoint, params)

        # build invoice object
        invoice = Invoice(self._client)
        return util.convert_to_object(invoice, response['body'])

class Email(InvoicedObject):
    pass


class Estimate(CreateableObject, DeleteableObject, ListableObject,
               UpdateableObject):

    def send(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/emails"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build email objects
        email = Email(self._client)
        return util.build_objects(email, response['body'])

    def attachments(self, **params):
        response = self._client.request('GET',
                                        self.endpoint()+"/attachments",
                                        params)

        # ensure each attachment has an ID
        body = response['body']
        for attachment in body:
            if 'id' not in attachment:
                attachment['id'] = attachment['file']['id']

        # build attachment objects
        attachment = Attachment(self._client)
        attachments = util.build_objects(attachment, body)

        # store the metadata from the list operation
        metadata = List(response['headers']['link'],
                        response['headers']['x-total-count'])

        return attachments, metadata

    def invoice(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/invoice"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build invoice object
        invoice = Invoice(self._client)
        return util.convert_to_object(invoice, response['body'])

    def void(self):
        endpoint = self.endpoint()+'/void'
        response = self._client.request('POST', endpoint)

        # update the local values with the response
        self.refresh_from(response['body'])

        return response['code'] == 200


class Event(ListableObject):
    pass


class File(CreateableObject, DeleteableObject):
    pass


class Invoice(CreateableObject, DeleteableObject, ListableObject,
              UpdateableObject):

    def send(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/emails"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build email objects
        email = Email(self._client)
        return util.build_objects(email, response['body'])

    def send_sms(self, idempotency_key=None, **opts):
        endpoint = self.endpoint()+"/text_messages"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, {}, opts)

        # build text message objects
        text_message = TextMessage(self._client)
        return util.build_objects(text_message, response['body'])

    def send_letter(self, idempotency_key=None, **opts):
        endpoint = self.endpoint()+"/letters"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, {}, opts)

        # build letter objects
        letter = Letter(self._client)
        return util.build_objects(letter, response['body'])

    def pay(self, idempotency_key=None, **opts):
        endpoint = self.endpoint()+"/pay"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, {}, opts)

        # update the local values with the response
        self.refresh_from(response['body'])

        return response['code'] == 200

    def attachments(self, **params):
        response = self._client.request('GET',
                                        self.endpoint()+"/attachments",
                                        params)

        # ensure each attachment has an ID
        body = response['body']
        for attachment in body:
            if 'id' not in attachment:
                attachment['id'] = attachment['file']['id']

        # build attachment objects
        attachment = Attachment(self._client)
        attachments = util.build_objects(attachment, body)

        # store the metadata from the list operation
        metadata = List(response['headers']['link'],
                        response['headers']['x-total-count'])

        return attachments, metadata

    def notes(self):
        note = Note(self._client)
        note.set_endpoint_base(self.endpoint())

        return note

    def payment_plan(self):
        paymentPlan = PaymentPlan(self._client)
        paymentPlan.set_endpoint_base(self.endpoint())

        return paymentPlan

    def void(self):
        endpoint = self.endpoint()+'/void'
        response = self._client.request('POST', endpoint)

        # update the local values with the response
        self.refresh_from(response['body'])

        return response['code'] == 200

class Letter(InvoicedObject):
    pass

class LineItem(CreateableObject, DeleteableObject, ListableObject,
               UpdateableObject):
    pass

class Note(CreateableObject, DeleteableObject, ListableObject,
           UpdateableObject):
    pass

class PaymentPlan(DeleteableObject):

    def __init__(self, client, id=None, values={}):
        super().__init__(client, id, values)

        self._endpoint = '/payment_plan'

    def create(self, idempotency_key=None, **params):
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('PUT', self.endpoint(), params, opts)

        return util.convert_to_object(self, response['body'])

    def retrieve(self, params={}):
        response = self._client.request('GET',
                                        self.endpoint(),
                                        params)

        return util.convert_to_object(self, response['body'])

    def cancel(self):
        return self.delete()

class PaymentSource(CreateableObject, ListableObject):
    pass

class Plan(CreateableObject, DeleteableObject, ListableObject,
           UpdateableObject):
    pass

class Subscription(CreateableObject, DeleteableObject, ListableObject,
                   UpdateableObject):

    def cancel(self):
        return self.delete()

    def preview(self, **params):
        self._endpoint = "/subscriptions/preview"
        repsonse = self._client.request('POST', self.endpoint(), params)

        return repsonse['body']

class Task(CreateableObject, DeleteableObject, ListableObject,
           UpdateableObject):
    pass

class TaxRate(CreateableObject, DeleteableObject, ListableObject,
              UpdateableObject):
    pass

class TextMessage(InvoicedObject):
    pass

class Transaction(CreateableObject, DeleteableObject, ListableObject,
                  UpdateableObject):

    def send(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/emails"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        # build email objects
        email = Email(self._client)
        return util.build_objects(email, response['body'])

    def refund(self, idempotency_key=None, **params):
        endpoint = self.endpoint()+"/refunds"
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        return util.convert_to_object(self, response['body'])

    def initiate_charge(self, idempotency_key=None, **params):
        endpoint = self._endpoint='/charges'
        opts = {'idempotency_key': idempotency_key}
        response = self._client.request('POST', endpoint, params, opts)

        return util.convert_to_object(self, response['body'])